import com.interfaceware.chameleon.*;
import PrefixTest.*;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.util.Timer;
import java.util.TimerTask;

/**
 * Class: ClientExample
 * 
 * Description:
 * 
 * The ClientExample is an example of handling the client networking events that
 * are dispatched from the operating system.
 */
class ClientExample {

    class ClientLog implements Log {
        /*
         * A real HL7 interface should have some type of system of logging into log
         * files.  Iguana, for instance, has a very nice logging system that allows 
         * administrators to drill into and browse processed messages and errors.  
         * How you implement this logging is up to your requirements for run time 
         * management of your HL7 interface.
         * 
         * For this example, we will log to the Java console.
         */
        public void println(String Line) {
            System.out.println(Line);
        }
    }

    class MyListener implements LlpClientListener {
        /*
         * The connected event is very important. One common mistake that 
         * developers who are new to socket programming often make is to try and 
         * invoke the connect method and then immediately send data via the socket.
         * This is not possible to do until the socket is actually connected.
         */
        public void connected() {
            printSeparator();
            println("Client is connected to " + cachedHost + " on port " + cachedPort);
            mainMenu();
        }

        /*
         * This event is fired when we receive a TCP/IP error on the connection.
         * This will always terminate the connection so your real application will 
         * need logic to handle this event in determining if and when it should 
         * attempt to re-connect to the other host.
         *
         * One of the advantages of using Iguana instead of Chameleon is that
         * Iguana will handle this for you.
         */
        public void errorReceived(String ErrorMessage, int ErrorCode) {
            printSeparator();
            println("Client error:");
            println(ErrorMessage);
            if(ackTimer != null) {
                ackTimer.cancel();
            }
            try {
                if(client.getIsConnected()) client.disconnect();
                client.stopMessageLoop();
            } catch(Exception e) {
                println(e.toString());
            }    
        }

        /* This event is triggered when the client socket receives an LLP 
         * wrapped HL7 message.  The envelope header and trailer characters will
         * have been stripped off the Message before being passed to your 
         * application in this event.
         * 
         * Most of the time the message received is an ACK message.  However if you
         * are implementing a lab query interface it might be a lab result message.
         * For this example we are assuming it is an ACK message and just printing 
         * it to the console.
         */
        public void messageReceived(String Message) {
            /*
             * For demonstration purposes, we assume any incoming message
             * is an ACK/NACK message.
             */
            if (ackTimer != null) {
                ackTimer.cancel();
            }

            printSeparator();
            println("Message received:");

            try {
                /*
                * The makeMessagePrintable function converts \r characters
                * (which are usually the segment delimiters of an HL7 message)
                * into the string given in the second argument of 
                * MakeMessagePrintable. Since \r (carriage return) only places
                * the cursor back to the beginning of the console line, we 
                * replace it with \n (line feed), which appends a new line to
                * the console.
                * 
                * The third argument is the maximum length of the message 
                * before we truncate it. If logging the message consumes too 
                * much processing time, we may specify a maximum length to 
                * prevent the Client application from hanging.
                */
                println(client.makeMessagePrintable(Message,"\n",-1));
            } catch (Exception e) {
                println(e.toString());
            }
            mainMenu();
        }

        /*
         * This event is generated when data which is not within the LLP header and
         * trailer characters is received. This should be a very rare event and 
         * usually indicates a problem with the implementation of the peer host 
         * that is sending HL7 back to your client.
         */ 
        public void dataIgnored(String Data) {
            printSeparator();
            println("Ignored data outside of LLP connection:");
            println(Data);
            mainMenu();
        }

        /*
         * This event is fired when a TCP/IP disconnection completes.
         */
        public void closed() {
            printSeparator();
            println("Client disconnected");
            if (ackTimer != null) {
                ackTimer.cancel();
            }
            try {
                client.stopMessageLoop();
            } catch(Exception e) {
                println(e.toString());
            }
        }
    }

    LlpClient client;
    PrefixTestEngine clientEngine;
    String cachedHost;
    int cachedPort;
    BufferedReader stdin;
    Timer ackTimer;
    int ackTimerDelay;
    boolean waitingToQuit;
    
    /**
     * Main entry point.
     */
    static public void main(String[] Args) throws Exception {
        ClientExample Client = new ClientExample();
        Client.runClient();
    }

    /**
     * Constructor.
     */
    public ClientExample() throws Exception {
        client = new LlpClient();
        client.addListener(new MyListener());
        stdin = new BufferedReader(new InputStreamReader(System.in));
        cachedHost = "localhost";
        cachedPort = 5123;
        ackTimer = null;
        ackTimerDelay = 10000; /* 10 seconds */
        waitingToQuit = false;
    }

    public void runClient() {
        try {
            System.out.println("HL7 Client started");
            initializeEngine();

            /* 
             * runMessageLoop() is a blocking call to listen for client
             * networking events from the operating system, and then dispatch
             * these events to the listenerInstance.
             */ 
            while(!waitingToQuit) {
                mainMenu();
                if(!waitingToQuit) LlpClient.runMessageLoop();
            }
        } catch (Exception e) {
            System.out.println(e.toString());
        } finally {
            System.out.println("-----");
            System.out.println("HL7 Client terminated");
            System.exit(0);
        }
    }  

    /*
     * A real HL7 interface should have some type of system of logging into log
     * files.  Iguana for instance has a very nice logging system that allows 
     * administrators to drill into and browse processed messages and errors.  
     * How you implement this logging is up to your requirements for run time 
     * management of your HL7 interface.
     */
    public void println(String Line) {
        System.out.println(Line);
    }

    public void printSeparator() {
        println("-----");
    }

    public void printPrompt(String Prompt) {
        System.out.print(Prompt);
    }

    public void initializeEngine() throws ChameleonException {
        printSeparator();

        /*
         * The first step in making the generated PrefixTestEngine object useable 
         * is to load the message definition or 'vmd' file. This file contains
         * the information about the structure of the messages, tables, and 
         * data mappings that were defined in the Chameleon GUI. 
         *  
         * PrefixTestEngine is a generated engine which means that it has an interface
         * based on the Messages and Tables you define in your vmd file. The 
         * use of stub code is helpful since it means that the compiler can 
         * catch typos or spelling mistakes of fields at compile rather than 
         * run time. However, it is not the only way you can use Chameleon. It
         * is possible to use a non-stubcode interface for the Chameleon
         * engine which we call the 'dynamic interface'. Please review the 
         * manual documentation for more details if you need this functionality
         * - most customers find that the stub code interface works well for 
         * their applications.
         */
        clientEngine = new PrefixTestEngine();
        clientEngine.addLog(new ClientLog());
        println("Loading vmd files...");
        clientEngine.load("test_data.vmd");
        println("test_data.vmd loaded");
    }

    /*
     * Setting up the various menus is just part of mechanics of this example 
     * which is for teaching purposes only.  A real HL7 client would most 
     * likely be part of a backend service.
    */
    public void mainMenu() {
        try
        {
            printSeparator();
            println("Available commands:");
            println("0) Quit");
            println("1) Change Configuration");
            if (!client.getIsConnected()) {
                println("2) Connect");
            } else {
                println("2) Disconnect");
                println("3) Send Message");
            }

            printPrompt("Enter command number: ");
            char InputChar = stdin.readLine().charAt(0);
            switch (InputChar) {
            case '0':
                waitingToQuit = true;
                if (client.getIsConnected()) {
                    /*
                    * We can't terminate the message loop 
                    * until we are actually disconnected.
                    */
                    client.disconnect();
                } else {
                    client.stopMessageLoop();
                }
                break;
            case '1':
                configurationMenu();
                break;
            case '2':
                if (!client.getIsConnected()) {
                    connectPrompt();
                } else {
                    client.disconnect();
                }
                break;
            case '3':
                if(client.getIsConnected()) {
                    sendMenu();
                    break;
                }
            default:
                printSeparator();
                println("Invalid command number!");
                mainMenu();
            }
        } catch (Exception e) {
            printSeparator();
            println(e.toString());
            mainMenu();
        }
    }

    public void configurationMenu() throws Exception {
        printSeparator();
        println("Available configurations in test_data.vmd:");
        for (int ConfigIndex = 0; ConfigIndex < clientEngine.countOfConfig(); ConfigIndex++)
        {
            println(ConfigIndex + ") " + clientEngine.configName(ConfigIndex));
        }

        printPrompt("Enter configuration index (current: " + clientEngine.currentConfigIndex() + "): ");
        String ConfigInput = stdin.readLine();
        if (ConfigInput.length() > 0) {
            clientEngine.setCurrentConfigIndex(Integer.valueOf(ConfigInput).intValue());
        }

        printSeparator();
        println("Using " + clientEngine.configName(clientEngine.currentConfigIndex()) + " configuration");

        mainMenu();
    }

    public void connectPrompt() {
        try {
            printSeparator();
            printPrompt("Enter host address (default: " + cachedHost + "): ");
            String HostInput = stdin.readLine();
            if (HostInput.length() > 0) {
                cachedHost = HostInput;
            }

            printPrompt("Enter port (default: " + cachedPort + "): ");
            String PortInput = stdin.readLine();
            if (PortInput.length() > 0) {
                cachedPort = Integer.valueOf(PortInput).intValue();
            }

            client.connect(cachedHost, cachedPort);
        } catch (Exception e) {
            printSeparator();
            println(e.toString());
            mainMenu();
        }
    }

    public void sendMenu() throws Exception
    {
        printSeparator();
        println("Available messages:");
        for (int MessageIndex = 0; 
                 MessageIndex < clientEngine.countOfMessage(); 
                 MessageIndex++) {
            println(MessageIndex + ") " + clientEngine.messageName(MessageIndex));
        }

        printPrompt("Enter message index: ");
        int MessageIndex = Integer.parseInt(stdin.readLine());
        println("");
        String Message;
        switch (MessageIndex) {
        // Begin message generation calls (comment used as an anchor for STJ)
        case 0:
            Message = generateADT_A04();
            break;
        // End message generation calls (comment used as an anchor for STJ)

        default:
            throw new ChameleonException("Invalid message index!");
        }

        printSeparator();
        println("Sending message:");

        /*
         * The makeMessagePrintable function converts \r characters
         * (which are usually the segment delimiters of an HL7 message)
         * into the string given in the second argument of 
         * MakeMessagePrintable. Since \r (carriage return) only places
         * the cursor back to the beginning of the console line, we 
         * replace it with \n (line feed), which appends a new line to
         * the console.
         * 
         * The third argument is the maximum length of the message 
         * before we truncate it. If logging the message consumes too 
         * much processing time, we may specify a maximum length to 
         * prevent the Client application from hanging.
         */
        println(client.makeMessagePrintable(Message,"\n",-1));

        client.transmitMessage(Message);

        /*
         * We need to specify a maximum waiting period in case the 
         * listening server never responds with an ACK/NACK message.
         */  
        ackTimer = new Timer();
        ackTimer.schedule(new AckTimerTask(), ackTimerDelay);
    }

    class AckTimerTask extends TimerTask {
        public void run() {
            printSeparator();
            println("No ACK/NACK message was received within the specified time interval.");
            mainMenu();
        }
    }

    /*
     * These generated methods are examples of how an HL7 message can be 
     * created using Chameleon's generated stubcode objects. We are populating
     * these objects with dummy data. In a real application, you would need to
     * obtain this data from somewhere (e.g. a database).
     * 
     * It should be noted that each Message object corresponds to the message 
     * defined in the vmd file loaded into the clientEngine. These objects have
     * tables and sub tables with properties that you can populate the message
     * with.
     * It must be emphasized that this will be a mirror of the 'table grammar' 
     * for each message - i.e. the user defined definitions of each table and 
     * the fields in those tables as defined in the Chameleon GUI.
     *
     * Also keep in mind that this is using the static stubcode interface of 
     * Chameleon. For advanced applications, it is possible to use a dynamic 
     * interface to populate messages.  See the manual for more details.
     */
    // Begin message generation functions (comment used as an anchor for STJ)
    public String generateADT_A04() throws Exception {
        ADT_A04Message Message = new ADT_A04Message(clientEngine);
        Message.PatientInfo().addRow();
        Message.PatientInfo(0).setSetIDPid(0);
        Message.PatientInfo(0).setPatientNameFamilyNameSurnamerepeatsonlyfirstinstancemapped("PatientNameFamilyNameSurnamerepeatsonlyfirstinstancemapped Dummy Value");
        Message.PatientInfo(0).setPatientNameGivenNamerepeatsonlyfirstinstancemapped("PatientNameGivenNamerepeatsonlyfirstinstancemapped Dummy Value");
        Message.PatientInfo(0).setDatetimeOfBirthTime(ChameleonDateTime.now());
        Message.PatientInfo(0).setAdministrativeSex("AdministrativeSex Dummy Value");
        return Message.generateMessage();
    }
    // End message generation functions (comment used as an anchor for STJ)


}
