import com.interfaceware.chameleon.*;
import com.interfaceware.iguana.plugin.*;

/**
 * Class: DestinationPluginExample
 * 
 * Description:
 * 
 * The DestinationPluginExample is an example of an Iguana destination plugin
 * using Chameleon to parse received messages.
 */
class DestinationPluginExample {

    class DestinationPluginLog implements Log {
        /*
         * A real HL7 interface should have some type of system of logging into log
         * files.  Iguana, for instance, has a very nice logging system that allows 
         * administrators to drill into and browse processed messages and errors.  
         * How you implement this logging is up to your requirements for run time 
         * management of your HL7 interface.
         * 
         * For this example, we will log to the Java console, and we'll collect
         * log messages to send back to Iguana to be logged in Iguana's logging system.
         * See addInfoLogMessage below.
         */
        public void println(String Line) {
            System.out.println(Line);
            logString += Line;
            logString += "\n";
        }
    
        String logString;
    }
    
    class MyDestinationListener implements DestinationListener {

        /*
         * Message processing invoked by messageReceived and 
         * resubmissionPreviewMessageReceived event handlers.
         */
        public void processMessage
        (
            DestinationMessageReceivedEventInputs inputs,
            DestinationMessageReceivedEventOutputs outputs,
            boolean resubmissionPreview
        ) throws Exception
        {
            try {
                /*
                 * The resubmissionPreview flag indicates whether or not we're processing a resubmission 
                 * preview message.  When processing a resubmission preview message, the plugin should 
                 * simulate the processing that would happen for a regular message, but without causing 
                 * external side effects.  For example, no database commits should be made for resubmission
                 * preview messages.  In this example, regular message processing has no external side 
                 * effects, so we simply ignore the resubmissionPreview flag.
                 */

                println("");
                println("*** Message Received ***"); 
                println(inputs.getMessage().replace('\r', '\n'));
                
                /* 
                 * Notice that we 'manually' invoke the parseMessage method on the 
                 * parseEngine object. This will fire one of the generated onMessage 
                 * events which are defined in the DestinationPluginEngine class.
                 * 
                 * The plugin classes are not tightly coupled to the HL7Engine object. 
                 * You can use any kind of transport you like with  Chameleon, including 
                 * your own transports or third party components.
                 * 
                 * In this application we do not block on events - this is because the
                 * plugin components are asynchronous. You can write synchronous 
                 * applications using Chameleon but you would need to use another 
                 * transport mechanism other than the plugin components supplied here.
                 */
                log.logString = "";
                parseEngine.parseMessage(inputs.getMessage(), null);

                // Tell Iguana that the message has been successfully processed.
                outputs.setMessageProcessedFlag(true);

            } catch (Exception e) {
                e.printStackTrace();

                // Tell Iguana that the message was not successfully processed.
                outputs.setMessageProcessedFlag(false);

                // Add an error-level log message to Iguana's logs describing the failure.
                outputs.addErrorLogMessage(e.toString());

                /* You can use ResendMessageFlag and ResendMessageDelay to request that Iguana
                 * resend the same message again after a delay period.  This is useful in situations
                 * where the message processing failure was due to a time sensitive cause, and
                 * message processing might succeed at a later time.
                 *
                 * outputs.setResendMessageFlag( true );
                 * outputs.setResendMessageDelay( 1000 ); // Resend after 1 second
                 *
                 * You can use StopChannelFlag, StopChannelErrorFlag, and StopChannelReason to 
                 * stop the Iguana  channel with or without an error flag, and provide a reason 
                 * for the stop request.
                 * 
                 * outputs.setStopChannelFlag(true);
                 * outputs.setStopChannelErrorFlag(true);
                 * outputs.setStopChannelReason("An error occurred.");
                 */
            }

            /* You can send additional error/warning/info/debug log messages to Iguana.
             * In this case we simply send an info-level log message containing the 
             * Chameleon engine's log output.
             */
            if (log.logString.length() > 0) {
              outputs.addInfoLogMessage(log.logString);
            } 
        }

        /* 
         * This event is triggered when a message is received from Iguana.
         */
        public void messageReceived
        (
            DestinationMessageReceivedEventInputs inputs,
            DestinationMessageReceivedEventOutputs outputs
        )
        {
            // Note that exceptions can never be thrown outside the handler.
            try {

                processMessage(inputs, outputs, false);

            } catch (Exception e) {
                e.printStackTrace();
            }
        }

        /* 
         * This event is triggered when a resubmission preview message is received from Iguana.
         */
        public void resubmissionPreviewMessageReceived
        (
            DestinationMessageReceivedEventInputs inputs,
            DestinationMessageReceivedEventOutputs outputs
        )
        {
            // Note that exceptions can never be thrown outside the handler.
            try {
                /*
                 * This handler should simulate the processing that would be performed by the
                 * messageReceived event handler without causing external side effects.
                 * For example, database changes should not be made for resubmission preview messages.
                 */

                processMessage(inputs, outputs, true);

            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    
        /* 
         * This event is received when the plugin disconnects from Iguana.
         * See also the use of waitForDisconnect() below.
         */
        public void disconnected() {
            // Note that exceptions can never be thrown outside the handler.
            try {
                println("Destination disconnected.");
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    }

    Destination destination;
    DestinationPluginEngine parseEngine;
    DestinationPluginLog log;

    /**
     * Main entry point.
     */
    static public void main(String[] args) throws Exception {
        try {
            if (args.length < 3) {
                throw new Exception("Invalid command line arguments.");
            }

            String hostname  = args[0];
            int    port      = Integer.parseInt(args[1]);
            String channel   = args[2];

            DestinationPluginExample destinationPluginExample = new DestinationPluginExample();
            destinationPluginExample.destination.connect(hostname, port, channel);

            // Wait for the plugin to disconnect.
            // Could also use waitForDisconnectWithTimeout() here.
            destinationPluginExample.destination.waitForDisconnect();

            // Print the reason for disconnection.
            DisconnectReason disconnectReason = destinationPluginExample.destination.getDisconnectReason();
            System.out.println(disconnectReason.getDescription());

        } catch (Exception e) {
            e.printStackTrace();
            System.out.println();
            System.out.println("This is an Iguana destination plugin.  It runs as a command line application");
            System.out.println("which connects via TCP/IP to the plugin communication server within Iguana.");
            System.out.println("It will be necessary to run a channel in Iguana and set up a channel with");
            System.out.println("a To Plugin component for this to connect to.  See:");
            System.out.println();
            System.out.println("   http://www.interfaceware.com/manual/plugin_setup_java.html");
            System.out.println();
            System.out.println("Once the channel is set up and running you can connect to it with this usage:");
            System.out.println();
            System.out.println("  DestinationPluginExample <iguana hostname> <plugin communication port> <channel name>");
            System.out.println();
            System.out.println("The plugin communication port is the port that Iguana listens on for plugin");
            System.out.println("communication.");
        }
    }

    /**
     * Constructor.
     */
    public DestinationPluginExample() throws Exception {
        println("Destination plugin started");

        destination = new Destination();
        destination.addListener(new MyDestinationListener());

        initializeEngine();

        /* 
         * Specifies the thread that will be run when ConsoleServer 
         * terminates. See the class definition for ShutdownThread below
         * for more details.
         */
        Runtime.getRuntime().addShutdownHook(new ShutdownThread());
    }  

    /* 
     * The ShutdownThread will be run when the ConsoleServer 
     * terminates. This thread is guaranteed to run no matter how the
     * ConsoleServer is terminated.
     * 
     * This ensures that when ConsoleServer terminates, we disconnect from
     * Iguana.
     */
    class ShutdownThread extends Thread	{
        public void run() {
            try {
                destination.disconnect();
            } catch (Exception e) {
                e.printStackTrace();
            } finally {
                System.out.println("Destination plugin terminated");
            }
        }
    }

    public void println(String Line) {
        System.out.println(Line);
    }

    public void initializeEngine() throws Exception {
        /*
         * The first step in making the generated parsing engine component 
         * useable is to load the message definition or 'vmd' file. This file 
         * contains all the definitions for the messages, how they are 
         * identified and how they should be mapped into the generated data 
         * objects based on the tables that you define in the Chameleon GUI.
         *  
         * DestinationPluginEngine subclasses from HL7Engine since it implements the 
         * actual handling for each particular type of message. HL7Engine is a
         * generated engine which means that it has an interface based on the 
         * Messages and Tables you define in your vmd file.  The use of stub 
         * code is helpful since it means that the compiler can catch typos or
         * spelling mistakes of fields at compile rather than run time. 
         * However, it is not the only way you can use Chameleon. It is 
         * possible to use a non-stubcode interface for the Chameleon parsing
         * engine which we call the 'dynamic interface'. Please review the 
         * manual documentation for more details if you need this functionality
         * - most customers find that the stub code interface works well for 
         * their applications.
         */
        println("Loading vmd files...");
        log = new DestinationPluginLog();
        parseEngine = new DestinationPluginEngine();
	parseEngine.addLog(log);
        parseEngine.load("test_data.vmd");
        parseEngine.setCurrentConfigIndex(0);
        println("test_data.vmd loaded");
    }
}
